@@ -1,50 +1,25 @@
 """
-TkinterWeb v3.23
-This is a wrapper for the Tkhtml3 widget from http://tkhtml.tcl.tk/tkhtml.html, 
-which displays styled HTML documents in Tkinter.
+Widgets that expand on the functionality of the basic bindings
+by adding scrolling, file loading, and many other convenience functions
 
-Copyright (c) 2024 Andereoo
-
-Permission is hereby granted, free of charge, to any person obtaining a copy
-of this software and associated documentation files (the "Software"), to deal
-in the Software without restriction, including without limitation the rights
-to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
-copies of the Software, and to permit persons to whom the Software is
-furnished to do so, subject to the following conditions:
-
-The above copyright notice and this permission notice shall be included in all
-copies or substantial portions of the Software.
-
-THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
-IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
-FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
-AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
-LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
-OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
-SOFTWARE.
+Copyright (c) 2025 Andereoo
 """
 
+import platform
+from urllib.parse import urldefrag, urlparse
+
 from bindings import TkinterWeb
-from utilities import (AutoScrollbar, StoppableThread, cachedownload, download,
-                   notifier, currentpath, threadname)
-from imageutils import newimage
-import platform
+from utilities import (WORKING_DIR, BUILTINPAGES, AutoScrollbar, StoppableThread, cachedownload, download,
+                       notifier, threadname, __version__)
+from imageutils import ImageLabel
+from dom import TkwDocumentObjectModel
 
-try:
-    from urllib.parse import urlparse, urldefrag
-except ImportError:
-    from urlparse import urlparse, urldefrag
-
-try:
-    import tkinter as tk
-    from tkinter import ttk
-except ImportError:
-    import Tkinter as tk
-    import ttk
+import tkinter as tk
+from tkinter import ttk
 
 
 class HtmlFrame(ttk.Frame):
-    def __init__(self, master, messages_enabled=True, vertical_scrollbar="auto", horizontal_scrollbar=False, scroll_overflow=None, **kw):
+    def __init__(self, master, messages_enabled=True, vertical_scrollbar="auto", horizontal_scrollbar=False, overflow_scroll_frame=None, **kw):
         ttk.Frame.__init__(self, master, **kw)
 
         if messages_enabled:
@@ -53,68 +28,59 @@ class HtmlFrame(ttk.Frame):
             self.message_func = message_func = lambda message: None
 
         # setup scrollbars and HTML widget
-        self.html = html = TkinterWeb(self, message_func, HtmlFrame)
+        self.html = html = TkinterWeb(self, message_func, HtmlFrame, self.manage_vsb, self.manage_hsb, overflow_scroll_frame)
         html.grid(row=0, column=0, sticky=tk.NSEW)
 
-        if vertical_scrollbar:
-            if vertical_scrollbar == "auto":
-                self.vsb = vsb = AutoScrollbar(
-                    self, orient=tk.VERTICAL, command=html.yview)
-            else:
-                self.vsb = vsb = ttk.Scrollbar(
-                    self, orient=tk.VERTICAL, command=html.yview)
+        self.document = TkwDocumentObjectModel(html)
 
-            vsb.bind("<Enter>", html.on_leave)
-            vsb.bind("<MouseWheel>", self.scroll)
-            vsb.bind("<Button-4>", self.scroll_x11)
-            vsb.bind("<Button-5>", self.scroll_x11)
-            html.bind("<Button-4>", self.overflow_scroll_x11)
-            html.bind("<Button-5>", self.overflow_scroll_x11)
-            self.bind_class("{0}.document".format(html),
-                            "<MouseWheel>", self.scroll)
-            self.bind_class(html.scrollable_node_tag,
-                            "<MouseWheel>", self.scroll)
-            self.bind_class(html.scrollable_node_tag,
-                            "<Button-4>", self.scroll_x11)
-            self.bind_class(html.scrollable_node_tag,
-                            "<Button-5>", self.scroll_x11)
+        self.background = ttk.Style().lookup('TFrame', 'background')
 
-            html.configure(yscrollcommand=vsb.set)
-            vsb.grid(row=0, column=1, sticky=tk.NSEW)
+        self.image_label = ImageLabel(html, self.on_error, anchor='center')
 
-        if horizontal_scrollbar:
-            if horizontal_scrollbar == "auto":
-                self.hsb = hsb = AutoScrollbar(
-                    self, orient=tk.HORIZONTAL, command=html.xview)
-            else:
-                self.hsb = hsb = ttk.Scrollbar(
-                    self, orient=tk.HORIZONTAL, command=html.xview)
+        self.hsb = hsb = AutoScrollbar(self, orient=tk.HORIZONTAL, command=html.xview)
+        self.hsb_type = None
+        self.original_hsb_type = horizontal_scrollbar
+        hsb.bind("<Enter>", self.html.on_leave)
+        html.configure(xscrollcommand=hsb.set)
+        hsb.grid(row=1, column=0, sticky="nsew")
+        self.manage_hsb(horizontal_scrollbar)
 
-            hsb.bind("<Enter>", html.on_leave)
+        self.vsb = vsb = AutoScrollbar(self, orient=tk.VERTICAL, command=self.html.yview)
+        self.vsb_type = None
+        self.original_vsb_type = vertical_scrollbar
+        vsb.bind("<Enter>", self.html.on_leave)
+        html.configure(yscrollcommand=vsb.set)
+        vsb.grid(row=0, column=1, sticky="nsew")
+        self.manage_vsb(vertical_scrollbar)
 
-            html.configure(xscrollcommand=hsb.set)
-            hsb.grid(row=1, column=0, sticky=tk.NSEW)
+        # for some reason, binding to Html only works on Linux and binding to html.document only works on Windows
+        # html.document only applies to the document it is bound to (which makes things easy)
+        # Html fires on all documents (i.e. <iframe> elements), so it has to be handled slightly differently
+        if not overflow_scroll_frame:
+            self.bind_class("Html", "<Button-4>", self.html.scroll_x11)
+            self.bind_class("Html", "<Button-5>", self.html.scroll_x11)
+        self.bind_class(f"{self.html}.document", "<MouseWheel>", self.html.scroll)
+
+        self.bind_class(self.html.scrollable_node_tag, "<Button-4>", lambda event, widget=html: self.html.scroll_x11(event, widget))
+        self.bind_class(self.html.scrollable_node_tag, "<Button-5>", lambda event, widget=html: self.html.scroll_x11(event, widget))
+        self.bind_class(self.html.scrollable_node_tag, "<MouseWheel>", self.html.scroll)
+
+        vsb.bind("<Button-4>", lambda event, widget=html: self.html.scroll_x11(event, widget))
+        vsb.bind("<Button-5>", lambda event, widget=html: self.html.scroll_x11(event, widget))
+        vsb.bind("<MouseWheel>", self.html.scroll)
 
         self.bind("<Leave>", html.on_leave)
         self.bind("<Enter>", html.on_mouse_motion)
 
         # state and settings variables
         self.master = master
-        self.scroll_overflow = scroll_overflow
         self.current_url = ""
         self.cursor = ""
         self.accumulated_styles = []
         self.waiting_for_reset = False
-        self.image_count = 0
         self.image = None
         self.thread_in_progress = None
-        self.broken_page_msg = """<html>
-                                    <head><title>Error 404</title></head>
-                                    <body style="text-align:center;">
-                                        <h2>Oops.</h2><p></p>
-                                        <h3>The page you've requested could not be found.</h3>
-                                    </body>
-                                  </html>"""
+        self.broken_page_msg = ""
         html.cursor_change_func = self.change_cursor
         html.link_click_func = self.load_url
         html.form_submit_func = self.load_form_data
@@ -123,29 +89,54 @@ class HtmlFrame(ttk.Frame):
         self.html.done_loading_func = self.done_loading
 
         self.message_func(
-            "Welcome to TkinterWeb 3.23! \nhttps://github.com/Andereoo/TkinterWeb")
+            f"Welcome to TkinterWeb v{__version__}! \nhttps://github.com/Andereoo/TkinterWeb")
 
         self.message_func(
-            "Debugging messages are enabled. \nUse the parameter `messages_enabled = False` when calling HtmlFrame() to disable these messages.")
+            "Debugging messages are enabled \nUse the parameter `messages_enabled = False` when calling HtmlFrame() or HtmlLabel() to disable these messages")
 
         self.columnconfigure(0, weight=1)
         self.rowconfigure(0, weight=1)
 
-        #Redirected commands
+        # redirected commands
         self.select_all = self.html.select_all
         self.bind = self.html.bind
-        self.set_zoom = self.html.set_zoom
-        self.get_zoom = self.html.get_zoom
         self.set_fontscale = self.html.set_fontscale
         self.get_fontscale = self.html.get_fontscale
         self.set_parsemode = self.html.set_parsemode
         self.get_parsemode = self.html.get_parsemode
         self.resolve_url = self.html.resolve_url
+        self.get_zoom = self.html.get_zoom
+
+        self.replace_widget = self.html.replace_widget
+        self.replace_element = self.html.replace_element
+        self.remove_widget = self.html.remove_widget
 
         self.yview = self.html.yview
         self.yview_moveto = self.html.yview_moveto
         self.yview_scroll = self.html.yview_scroll
-        
+
+    def manage_vsb(self, allow=None):
+        "Show or hide the scrollbars"
+        if allow == None:
+            allow = self.original_vsb_type
+        if allow == "auto":
+            allow = 2 
+        if self.vsb_type != allow:
+            self.vsb.set_type(allow)
+            self.vsb_type = allow
+        return allow
+    
+    def manage_hsb(self, allow=None):
+        "Show or hide the scrollbars"
+        if allow == None:
+            allow = self.original_hsb_type
+        if allow == "auto":
+            allow = 2
+        if self.hsb_type != allow:
+            self.hsb.set_type(allow)
+            self.hsb_type = allow
+        return allow
+
     def yview_toelement(self, selector, index=0):
         "Find an element that matches a given CSS selectors and scroll to it"
         nodes = self.html.search(selector)
@@ -155,13 +146,13 @@ class HtmlFrame(ttk.Frame):
             except IndexError:
                 pass
 
-    def load_website(self, website_url, decode=None, force=False, insecure=False):
+    def load_website(self, website_url, decode=None, force=False, insecure=None):
         "Load a website from the specified URL"
         if (not website_url.startswith("https://")) and (not website_url.startswith("http://")) and (not website_url.startswith("about:")):
             website_url = "http://" + str(website_url)
         self.load_url(website_url, decode, force, insecure)
 
-    def load_file(self, file_url, decode=None, force=False, insecure=False):
+    def load_file(self, file_url, decode=None, force=False, insecure=None):
         "Load a locally stored file from the specified path"
         if not file_url.startswith("file://"):
             if platform.system() == "Windows" and not file_url.startswith("/"):
@@ -170,18 +161,22 @@ class HtmlFrame(ttk.Frame):
                 file_url = "file://" + str(file_url)
         self.load_url(file_url, decode, force, insecure)
 
-    def load_url(self, url, decode=None, force=False, insecure=False):
+    def load_url(self, url, decode=None, force=False, insecure=None):
         """Load a website (https:// or http://) or a file (file://) from the specified URL.
         We use threading here to prevent the GUI from freezing while fetching the website.
-        Technically Tkinter isn't threadsafe and will crash when doing this, but under certain circumstances we can get away with it.
-        As long as we do not use the .join() method and no errors are raised in the mainthread, we should be okay.
-        """
+        Technically Tkinter isn't threadsafe, 
+        but as long as we do not use the .join() method and no errors are raised in the mainthread, we can get away with it.
+        """                        
+        if url in BUILTINPAGES:
+            self.load_html(BUILTINPAGES[url].format(self.background, "", "No file selected"))
+            return
+
         self.waiting_for_reset = True
 
-        #Workaround for Bug #40, where urllib.urljoin constructs improperly formatted urls on Linux when url starts with file:///
+        # workaround for Bug #40, where urllib.urljoin constructs improperly formatted urls on Linux when url starts with file:///
         if not url.startswith("file://///"):
             url = url.replace("file:////", "file:///")
-                
+
         if self.thread_in_progress:
             self.thread_in_progress.stop()
         if self.html.max_thread_count >= 1:
@@ -192,23 +187,29 @@ class HtmlFrame(ttk.Frame):
         else:
             self.continue_loading(url, decode=decode, force=force, insecure=insecure)
 
-    def load_form_data(self, url, data, method="GET", decode=None):
+    def load_form_data(self, url, data, method="GET", decode=None, insecure=None):
         "Load a webpage using form data"
         if self.thread_in_progress:
             self.thread_in_progress.stop()
         if self.html.max_thread_count >= 1:
             thread = StoppableThread(
-                target=self.continue_loading, args=(url, data, method, decode))
+                target=self.continue_loading, args=(url, data, method, decode), kwargs={"insecure": insecure})
             self.thread_in_progress = thread
             thread.start()
         else:
             self.continue_loading(url, data, method, decode)
 
-    def continue_loading(self, url, data="", method="GET", decode=None, force=False, insecure=False):
+    def continue_loading(self, url, data="", method="GET", decode=None, force=False, insecure=None):
         "Finish loading urls and handle URI fragments"
+        if insecure == None:
+            insecure = self.html.insecure_https
+        else:
+            self.html.insecure_https = insecure
+        code = 404
 
         self.html.downloading_resource_func()
         self.url_change_func(url)
+        
         try:
             method = method.upper()
             parsed = urlparse(url)
@@ -216,29 +217,48 @@ class HtmlFrame(ttk.Frame):
             if method == "GET":
                 url = str(url) + str(data)
 
-            # if url is different than the current one, load the new site.
+            # if url is different than the current one, load the new site
             if force or (method == "POST") or (self.skim(urldefrag(url)[0]) != self.skim(urldefrag(self.current_url)[0])):
-                self.message_func("Connecting to {0}.".format(parsed.netloc))
+                view_source = False
+                if url.startswith("view-source:"):
+                    view_source = True
+                    url = url.replace("view-source:", "")
+                    parsed = urlparse(url)
+                self.message_func("Connecting to {0}".format(parsed.netloc))
                 if insecure:
                     self.message_func("WARNGING: Using insecure HTTPS session")
                 if (parsed.scheme == "file") or (not self.html.caches_enabled):
-                    data, newurl, filetype = download(
+                    data, newurl, filetype, code = download(
                         url, data, method, decode, insecure)
                 else:
-                    data, newurl, filetype = cachedownload(
+                    data, newurl, filetype, code = cachedownload(
                         url, data, method, decode, insecure)
+                self.message_func(f"Successfully connected to {parsed.netloc}")
                 if threadname().isrunning():
-                    self.url_change_func(newurl)
-                    if "image" in filetype:
-                        image, error = newimage(data, "_htmlframe_img_{}_{}_".format(id(self), self.image_count), filetype, self.html.image_inversion_enabled)
-                        if error:
-                            self.html.image_setup_func(url, False)
+                    if view_source:
+                        newurl = "view-source:"+newurl
+                        self.url_change_func(newurl)
+                        data = str(data).replace("<","&lt;").replace(">", "&gt;")
+                        data = data.splitlines()
+                        length = int(len(str(len(data))))
+                        if len(data) > 1:
+                            data = "</code><br><code>".join(data)
+                            data = data.rsplit("<br><code>", 1)[0]
+                            data = data.split("</code><br>", 1)[1]
                         else:
-                            self.html.image_setup_func(url, True)
-                        self.load_html("<img style='max-width:100%' src='replace:{}'></img".format(image))
-                        self.image_count += 1
-                        self.image = image
+                            data = "".join(data)
+                        self.load_html(BUILTINPAGES["about:view-source"].format(self.background, length*9, data), newurl)
+                    elif "image" in filetype:
+                        self.url_change_func(newurl)
+                        # arguably it's much easier to simply pass the image to the HTML engine
+                        # but I like centered content and it's not possible to vertically align an image in Tkhtml without explicitly setting its height
+                        # so instead we embed a label widget in the page and deal with page resizing and zooming ourselves
+                        # since we're loading an image it doesn't matter if it's a Label widget that doesn't support find_text, selection, etc.
+                        # its a bit hacky, but hey, it works and looks very snazzy
+                        self.load_html(BUILTINPAGES["about:image"].format(self.background, str(self.image_label)))
+                        self.after(0, self.image_label.load_image, newurl, data, filetype)
                     else:
+                        self.url_change_func(newurl)
                         self.load_html(data, newurl)
                     self.current_url = newurl
             else:
@@ -253,23 +273,42 @@ class HtmlFrame(ttk.Frame):
                 self.html.update()
                 try:
                     frag = ''.join(char for char in frag if char.isalnum() or char in ("-", "_"))
-                    node = self.html.search("[id=%s]" % frag)
+                    node = self.html.search(f"[id={frag}]")
                     if node:
                         self.html.yview(node)
                     else:
-                        node = self.html.search("[name=%s]" % frag)
+                        node = self.html.search(f"[name={frag}]")
                         if node:
                             self.html.yview(node)
                 except Exception:
                     pass
         except Exception as error:
-            self.message_func(
-                "An error has been encountered while loading {}: {}.".format(url, error))
-            self.load_html(self.broken_page_msg)
-            self.current_url = ""
+            self.on_error(url, error, code)
 
         self.thread_in_progress = None
 
+    def on_error(self, url, error, code):
+        self.message_func(
+            f"Error loading {url}: {error}")
+        if self.broken_page_msg:
+            self.load_html(self.broken_page_msg)
+        else:
+            self.load_html(BUILTINPAGES["about:error"].format(self.background, code))
+        self.current_url = ""
+        
+        if "CERTIFICATE_VERIFY_FAILED" in str(error):
+            python_version = ".".join(platform.python_version_tuple()[:2])
+            self.message_func(
+                f"Check that you are using the right url scheme. Some websites only support http.\n\
+This might also happen if your Python distribution does not come installed with website certificates.\n\
+This is a known Python bug on older MacOS systems. \
+Running something along the lines of \"/Applications/Python {python_version}/Install Certificates.command\" (with the qoutes) to install the missing certificates may do the trick.\n\
+Otherwise, use 'insecure=True' when loading a page to ignore website certificates.")
+
+    def set_zoom(self, multiplier):
+        self.html.set_zoom(multiplier)
+        self.image_label.set_zoom(float(multiplier))
+
     def skim(self, url):
         return url.replace("/", "")
 
@@ -367,16 +406,16 @@ class HtmlFrame(ttk.Frame):
         self.html.caches_enabled = enabled
 
     def enable_crash_prevention(self, enabled=True):
-        "Enable or disable extra crash prevention measures"
-        "Disabling this will remove all emojis, the noto color emoji font, and invalid rgb functions"
+        """Enable or disable extra crash prevention measures
+        Disabling this will remove all emojis, the noto color emoji font, and invalid rgb functions"""
         self.html.prevent_crashes = enabled
 
     def enable_dark_theme(self, enabled=True, invert_images=True):
-        "Enable or disable dark theme"
-        "This will cause page colours to be 'inverted' if enabled is set to True"
-        "This will also cause images to be inverted if 'invert_images' is also set to True"
-        if (enabled or invert_images): 
-            self.message_func("Warning: dark theme has been enabled. This feature is highly experimental and may cause freezes or crashes.")
+        """Enable or disable dark theme
+        This will cause page colours to be 'inverted' if enabled is set to True
+        This will also cause images to be inverted if 'invert_images' is also set to True"""
+        if (enabled or invert_images):
+            self.message_func("WARNING: dark theme has been enabled. This feature is highly experimental and may cause freezes or crashes.")
         self.html.dark_theme_enabled = enabled
         self.html.image_inversion_enabled = invert_images
         self.html.update_default_style()
@@ -409,7 +448,7 @@ class HtmlFrame(ttk.Frame):
 
     def get_current_link(self, resolve=True):
         "Convenience method for getting the url of the current hyperlink"
-        if self.get_currently_hovered_node_tag().lower() == "a": 
+        if self.get_currently_hovered_node_tag().lower() == "a":
             href = self.get_currently_hovered_node_attribute("href")
             if resolve:
                 return self.resolve_url(href)
@@ -441,8 +480,7 @@ class HtmlFrame(ttk.Frame):
         return text
 
     def get_currently_hovered_node_attribute(self, attribute):
-        """
-        Get the specified attribute of the HTML element the mouse pointer is currently over
+        """Get the specified attribute of the HTML element the mouse pointer is currently over
         For example, if the mouse is hovering over the element
         "<a href='example.com'></a>", calling "get_currently_hovered_node_attribute('href')" will return "example.com."
         """
@@ -457,61 +495,18 @@ class HtmlFrame(ttk.Frame):
         return attr
 
     def get_currently_selected_text(self):
-        "Get the text that is currently highlighted/selected."
+        "Get the text that is currently highlighted/selected"
         return self.html.get_selection()
 
-    def replace_widget(self, oldwidget, newwidget):
-        "Replace a stored widget"
-        self.html.replace_widget(oldwidget, newwidget)
-
-    def replace_element(self, cssselector, newwidget):
-        "Replace an HTML element with a widget"
-        self.html.replace_html(cssselector, newwidget)
-
-    def remove_widget(self, widget):
-        "Remove a stored widget"
-        self.html.remove_widget(widget)
-
-    def scroll(self, event):
-        "Handle mouse/touchpad scrolling"
-        yview = self.html.yview()
-        if self.scroll_overflow and yview[0] == 0 and event.delta > 0:
-            self.scroll_overflow.scroll(event)
-        elif self.scroll_overflow and yview[1] == 1 and event.delta < 0:
-            self.scroll_overflow.scroll(event)
-        elif platform.system() == "Darwin":
-            self.html.yview_scroll(int(-1*event.delta), "units")
-        else:
-            self.html.yview_scroll(int(-1*event.delta/30), "units")
-
-    def scroll_x11(self, event):
-        yview = self.html.yview()
-        if event.num == 4:
-            if self.scroll_overflow and yview[0] == 0:
-                self.scroll_overflow.scroll_x11(event)
-            else:
-                self.html.yview_scroll(-4, "units")
-        else:
-            if self.scroll_overflow and yview[1] == 1:
-                self.scroll_overflow.scroll_x11(event)
-            else:
-                self.html.yview_scroll(4, "units")
-    
-    def overflow_scroll_x11(self, event):
-        yview = self.html.yview()
-        if event.num == 4 and self.scroll_overflow and yview[0] == 0:
-            self.scroll_overflow.scroll_x11(event)
-        elif self.scroll_overflow and yview[1] == 1:
-            self.scroll_overflow.scroll_x11(event)
-
     def load_html(self, html_source, base_url=None):
         "Reset parser and send html code to it"
+        self.image_label.reset()
         self.html.reset()
         if not base_url:
-            path = currentpath(False)
+            path = WORKING_DIR
             if not path.startswith("/"):
-                path = "/{0}".format(path)
-            base_url = "file://{0}/".format(path)
+                path = f"/{path}"
+            base_url = f"file://{path}/"
         self.html.base_url = self.current_url = base_url
         self.html.parse(html_source)
 
@@ -525,12 +520,12 @@ class HtmlFrame(ttk.Frame):
             self.accumulated_styles = []
 
     def add_html(self, html_source):
-        "Parse HTML and add it to the end of the current document."
+        "Parse HTML and add it to the end of the current document"
         if not self.current_url:
-            path = currentpath(False)
+            path = WORKING_DIR
             if not path.startswith("/"):
-                path = "/{0}".format(path)
-            base_url = "file://{0}/".format(path)
+                path = f"/{path}"
+            base_url = f"file://{path}/"
             self.html.base_url = self.current_url = base_url
         self.html.parse(html_source)
 
@@ -549,7 +544,7 @@ class HtmlLabel(HtmlFrame):
         tags = list(self.html.bindtags())
         tags.remove("Html")
         self.html.bindtags(tags)
-        
+
         self.html.shrink(True)
 
-        self.load_html(text)
+        self.load_html(text)
\ No newline at end of file
